/*
 *  $Id: nameserver.c,v 1.13 2002/08/22 09:37:23 dreibh Exp $
 *
 * RSerPool implementation.
 *
 * Realized in co-operation between Siemens AG
 * and University of Essen, Institute of Computer Networking Technology.
 *
 * Acknowledgement
 * This work was partially funded by the Bundesministerium fr Bildung und
 * Forschung (BMBF) of the Federal Republic of Germany (Frderkennzeichen 01AK045).
 * The authors alone are responsible for the contents.
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * There are two mailinglists available at http://www.sctp.de/rserpool.html
 * which should be used for any discussion related to this implementation.
 *
 * Contact: rsplib-discussion@sctp.de
 *          dreibh@exp-math.uni-essen.de
 *
 * Purpose: Name Server
 *
 */


#include "tdtypes.h"
#include "utilities.h"
#include "netutilities.h"
#include "dispatcher.h"
#include "timer.h"

#include "pool.h"
#include "poolelement.h"
#include "poolnamespace.h"
#include "asapmessage.h"
#include "asapcreator.h"
#include "asapparser.h"
#include "asaperror.h"
#include "randomizer.h"
#include "breakdetector.h"

#include <ext_socket.h>
#include <net/if.h>
#include <sys/ioctl.h>



/* Exit immediately on Ctrl-C. No clean shutdown. */
// #define FAST_BREAK



#ifndef IPV6_JOIN_GROUP
#define IPV6_JOIN_GROUP IPV6_ADD_MEMBERSHIP
#endif
#ifndef IPV6_LEAVE_GROUP
#define IPV6_LEAVE_GROUP IPV6_DROP_MEMBERSHIP
#endif


struct ENRPSource
{
   int  FD;
   bool IsMulticast;
};

struct ENRPClient
{
   int                FD;
   char               Identifier[128];

   size_t             Registrations;
   struct NameServer* Owner;
   GList*             PoolElementList;
   struct Timer*      KeepAliveTimer;
   struct Timer*      TimeoutTimer;
};

struct NameServer
{
   GTree*                ClientTree;
   GTree*                SourceTree;
   struct PoolNamespace* Namespace;
   struct Dispatcher*    StateMachine;
   int                   MCastIPv4Socket;
   int                   ReplyIPv4Socket;
#ifdef HAVE_IPV6
   int                   MCastIPv6Socket;
   int                   ReplyIPv6Socket;
#endif

   card64                KeepAliveInterval;
   card64                KeepAliveTimeout;

   uint32_t              ServerID;
   bool                  IsMulticast;
};


#define ENRP_DEFAULT_KEEP_ALIVE_INTERVAL  2000000
#define ENRP_DEFAULT_KEEP_ALIVE_TIMEOUT   5000000



struct NameServer* nameServerNew();

void nameServerDelete(struct NameServer* nameServer);

void nameServerAddClient(struct NameServer* nameServer,
                         int                fd);

void nameServerRemoveClient(struct NameServer* nameServer,
                            struct ENRPClient* client);

bool nameServerAddSource(struct NameServer* nameServer,
                         int                sourceFD,
                         const bool         isMulticast);

bool nameServerJoinMulticastGroup(struct NameServer*       nameServer,
                                  struct sockaddr_storage* address,
                                  const char*              interface);

void nameServerRemoveSource(struct NameServer* nameServer,
                            int                sourceFD);

void nameServerMainLoop(struct NameServer* nameServer);

static void nameServerRemovePoolElement(struct NameServer*  nameServer,
                                        struct ENRPClient*  client,
                                        struct PoolElement* poolElement);

static void nameServerKeepAliveTimer(struct Dispatcher* dispatcher,
                                     struct Timer*      timer,
                                     void*              userData);

static void nameServerTimeoutTimer(struct Dispatcher* dispatcher,
                                   struct Timer*      timer,
                                   void*              userData);

static void nameServerHandleClientEvent(struct NameServer* nameServer,
                                        struct ENRPClient* client,
                                        int                fd);

static void nameServerSocketHandler(struct Dispatcher* dispatcher,
                                    int                fd,
                                    int                eventMask,
                                    void*              userData);



/* ###### Constructor #################################################### */
struct NameServer* nameServerNew()
{
   struct NameServer* nameServer = (struct NameServer*)malloc(sizeof(struct NameServer));
   if(nameServer != NULL) {
      nameServer->SourceTree        = g_tree_new(fdCompareFunc);
      nameServer->ClientTree        = g_tree_new(fdCompareFunc);
      nameServer->Namespace         = poolNamespaceNew();
      nameServer->StateMachine      = dispatcherNew(dispatcherDefaultLock,dispatcherDefaultUnlock,NULL);
      nameServer->MCastIPv4Socket   = ext_socket(AF_INET,  SOCK_DGRAM, IPPROTO_UDP);
      nameServer->ReplyIPv4Socket   = ext_socket(AF_INET,  SOCK_DGRAM, IPPROTO_UDP);
      if(nameServer->MCastIPv4Socket >= 0) {
         setNonBlocking(nameServer->MCastIPv4Socket);
      }
      if(nameServer->ReplyIPv4Socket >= 0) {
         setNonBlocking(nameServer->ReplyIPv4Socket);
      }
#ifdef HAVE_IPV6
      nameServer->MCastIPv6Socket   = ext_socket(AF_INET6, SOCK_DGRAM, IPPROTO_UDP);
      nameServer->ReplyIPv6Socket   = ext_socket(AF_INET6, SOCK_DGRAM, IPPROTO_UDP);
      if(nameServer->MCastIPv6Socket >= 0) {
         setNonBlocking(nameServer->MCastIPv6Socket);
      }
      if(nameServer->ReplyIPv6Socket >= 0) {
         setNonBlocking(nameServer->ReplyIPv6Socket);
      }
#endif
      nameServer->KeepAliveInterval = ENRP_DEFAULT_KEEP_ALIVE_INTERVAL;
      nameServer->KeepAliveTimeout  = ENRP_DEFAULT_KEEP_ALIVE_TIMEOUT;
      nameServer->IsMulticast       = false;
      nameServer->ServerID          = random32();
      if((nameServer->SourceTree   == NULL) ||
         (nameServer->ClientTree   == NULL) ||
         (nameServer->Namespace    == NULL) ||
         (nameServer->StateMachine == NULL)) {
         nameServerDelete(nameServer);
         nameServer = NULL;
      }
   }
   return(nameServer);
}


/* ###### Destructor ##################################################### */
void nameServerDelete(struct NameServer* nameServer)
{
   gpointer key;
   gpointer value;

   if(nameServer) {
      if(nameServer->ClientTree) {
         while(getFirstTreeElement(nameServer->ClientTree,&key,&value)) {
            nameServerRemoveClient(nameServer,(struct ENRPClient*)value);
         }
         g_tree_destroy(nameServer->ClientTree);
      }
      if(nameServer->SourceTree) {
         while(getFirstTreeElement(nameServer->SourceTree,&key,&value)) {
            nameServerRemoveSource(nameServer,(int)key);
         }
         g_tree_destroy(nameServer->SourceTree);
      }
      if(nameServer->Namespace) {
         poolNamespaceDelete(nameServer->Namespace);
         nameServer->Namespace = NULL;
      }
      if(nameServer->MCastIPv4Socket >= 0) {
         ext_close(nameServer->MCastIPv4Socket);
         nameServer->MCastIPv4Socket = -1;
      }
#ifdef HAVE_IPV6
      if(nameServer->MCastIPv6Socket >= 0) {
         ext_close(nameServer->MCastIPv6Socket);
         nameServer->MCastIPv6Socket = -1;
      }
#endif
      if(nameServer->StateMachine) {
         dispatcherDelete(nameServer->StateMachine);
         nameServer->StateMachine = NULL;
      }
      free(nameServer);
   }
}


/* ###### Add client ##################################################### */
void nameServerAddClient(struct NameServer* nameServer,
                         int                fd)
{
   struct sockaddr_storage name;
   socklen_t               namelen;

   struct ENRPClient* client = (struct ENRPClient*)malloc(sizeof(struct ENRPClient));
   if(client != NULL) {
      /* ====== Add client =============================================== */
      client->FD              = fd;
      client->PoolElementList = NULL;
      client->Owner           = nameServer;
      client->Registrations   = 0;
      client->KeepAliveTimer  = timerNew(nameServer->StateMachine,nameServerKeepAliveTimer,(void*)client);
      client->TimeoutTimer    = timerNew(nameServer->StateMachine,nameServerTimeoutTimer,(void*)client);
      if((client->KeepAliveTimer != NULL) && (client->TimeoutTimer != NULL)) {
         g_tree_insert(nameServer->ClientTree,(gpointer)client->FD,client);
         dispatcherAddFDCallback(nameServer->StateMachine,
                                 fd, FDCE_Read|FDCE_Exception,
                                 nameServerSocketHandler, (void*)nameServer);

         /* ====== Get information ========================================== */
         namelen = sizeof(name);
         if(ext_getpeername(fd,(struct sockaddr*)&name,&namelen) == 0) {
            if(address2string((struct sockaddr*)&name, (char*)&client->Identifier,
                              sizeof(client->Identifier), true) == false) {
               strcpy((char*)&client->Identifier,"(bad)");
            }
         }
         else {
            strcpy((char*)&client->Identifier,"?");
         }

         LOG_ACTION
         fprintf(stdlog,"Adding new client %s, socket #%d.\n",client->Identifier,client->FD);
         LOG_END
      }
      else {
         free(client);
         ext_close(fd);
      }
   }
   else {
      ext_close(fd);
   }
}


/* ###### Remove client ################################################## */
void nameServerRemoveClient(struct NameServer* nameServer,
                            struct ENRPClient* client)
{
   GList*              list;
   struct PoolElement* poolElement;

   if(client != NULL) {
      LOG_ACTION
      fprintf(stdlog,"Removing client %s, socket #%d.\n",client->Identifier,client->FD);
      LOG_END

      if(client->KeepAliveTimer) {
         timerDelete(client->KeepAliveTimer);
         client->KeepAliveTimer = NULL;
      }
      if(client->TimeoutTimer) {
         timerDelete(client->TimeoutTimer);
         client->TimeoutTimer = NULL;
      }

      list = g_list_first(client->PoolElementList);
      while(list != NULL) {
         poolElement = (struct PoolElement*)list->data;

         LOG_ACTION
         fprintf(stdlog,"Pool element #%Ld still registered -> removing.\n",(card64)poolElement->Handle);
         LOG_END

         nameServerRemovePoolElement(nameServer,client,poolElement);
         list = g_list_first(client->PoolElementList);
      }

      dispatcherRemoveFDCallback(nameServer->StateMachine,client->FD);
      g_tree_remove(nameServer->ClientTree,(gpointer)client->FD);
      ext_close(client->FD);
      free(client);
   }
}


/* ###### Add source ##################################################### */
bool nameServerAddSource(struct NameServer* nameServer,
                         int                sourceFD,
                         const bool         isMulticast)
{
   if(!isMulticast) {
      if(ext_listen(sourceFD,10) < 0) {
         LOG_ERROR
         logerror("listen() call failed");
         LOG_END
         return(false);
      }
   }

   if(nameServer != NULL) {
      struct ENRPSource* source = (struct ENRPSource*)malloc(sizeof(struct ENRPSource));
      if(source != NULL) {
         source->FD          = sourceFD;
         source->IsMulticast = isMulticast;
         if(isMulticast) {
            nameServer->IsMulticast = true;
         }

         g_tree_insert(nameServer->SourceTree,(gpointer)source->FD,source);

         dispatcherAddFDCallback(nameServer->StateMachine, sourceFD,
                                 FDCE_Read|FDCE_Exception,
                                 nameServerSocketHandler, (void*)nameServer);
         return(true);
      }
   }
   return(false);
}


/* ###### Remove source ################################################## */
void nameServerRemoveSource(struct NameServer* nameServer,
                            int                sourceFD)
{
   struct ENRPSource* source = (struct ENRPSource*)g_tree_lookup(nameServer->SourceTree,
                                                                 (gpointer)sourceFD);
   if(source != NULL) {
      dispatcherRemoveFDCallback(nameServer->StateMachine, sourceFD);
      g_tree_remove(nameServer->SourceTree,(gpointer)sourceFD);
      free(source);
   }
}


/* ###### Dump namespace ################################################# */
void nameServerDump(struct NameServer* nameServer)
{
   fputs("*************** Namespace Dump ***************\n",stdlog);
   printTimeStamp(stdlog);
   fputs("\n",stdlog);
   poolNamespacePrint(nameServer->Namespace,stdlog);
   fputs("**********************************************\n",stdlog);
}


/* ###### Main loop ###################################################### */
void nameServerMainLoop(struct NameServer* nameServer)
{
   if(nameServer) {
      for(;;) {
         dispatcherEventLoop(nameServer->StateMachine);
      }
   }
}


/* ###### Remove pool element ############################################ */
static void nameServerRemovePoolElement(struct NameServer*  nameServer,
                                        struct ENRPClient*  client,
                                        struct PoolElement* poolElement)
{
   struct Pool* pool = poolElement->OwnerPool;

   client->PoolElementList = g_list_remove(client->PoolElementList,poolElement);
   poolNamespaceRemovePoolElement(nameServer->Namespace,poolElement);
   poolRemovePoolElement(pool,poolElement);
   poolElementDelete(poolElement);
   poolElement = NULL;

   if(poolIsEmpty(pool)) {
      poolNamespaceRemovePool(nameServer->Namespace,pool);
      poolDelete(pool);
      pool = NULL;
   }
}


/* ###### Filter address array ########################################### */
static bool filterAddresses(struct PoolElement*            poolElement,
                            const struct sockaddr_storage* addressArray,
                            const size_t                   addresses)
{
   struct TransportAddress* transportAddress;
   GList*                   list;
   bool                     found;
   size_t                   i;
   size_t                   j;
   bool                     valid;

   valid = false;
   list = poolElement->TransportAddressList;
   while(list != NULL) {
      transportAddress = (struct TransportAddress*)list->data;
      for(i = 0;i < transportAddress->Addresses;i++) {
         found = false;
         for(j = 0;j < addresses;j++) {
            if(addresscmp((const struct sockaddr*)&transportAddress->AddressArray[i],
                          (const struct sockaddr*)&addressArray[j],
                          false) == 0) {
               found = true;
               break;
            }
         }

         if(!found) {
            LOG_VERBOSE3
            fputs("Rejected address ",stdlog);
            fputaddress((struct sockaddr*)&transportAddress->AddressArray[i],
                        false, stdlog);
            fputs("\n",stdlog);
            LOG_END

            if(i + 1 < transportAddress->Addresses) {
               for(j = i;j < transportAddress->Addresses - 1;j++) {
                  transportAddress->AddressArray[j] = transportAddress->AddressArray[j + 1];
               }
               i--;
            }
            transportAddress->Addresses--;
         }
         else {
            LOG_VERBOSE2
            fputs("Accepted address ",stdlog);
            fputaddress((struct sockaddr*)&transportAddress->AddressArray[i],
                        false, stdlog);
            fputs("\n",stdlog);
            LOG_END
         }
      }

      list = g_list_next(list);
      if(transportAddress->Addresses < 1) {
         LOG_VERBOSE3
         fputs("Removed empty TransportAddress structure\n",stdlog);
         LOG_END
         poolElementRemoveTransportAddress(poolElement,transportAddress);
         transportAddressDelete(transportAddress);
      }
      else {
         valid = true;
      }
   }

   return(valid);
}


/* ###### Handle registration request #################################### */
static bool nameServerRegistrationRequest(struct NameServer*  nameServer,
                                          struct ENRPClient*  client,
                                          int                 fd,
                                          struct ASAPMessage* message)
{
   struct PoolElement*      existing;
   struct Pool*             pool;
   bool                     poolAdded;
   struct sockaddr_storage* addressArray;
   int                      addresses;
   int                      i;
   bool                     addressesValid;

   LOG_ACTION
   fprintf(stdlog,"Registration Request: ");
   poolElementPrint(message->PoolElementPtr,stdlog);
   LOG_END


   message->Type  = AHT_REGISTRATION_RESPONSE;
   message->Error = AEC_OKAY;


   /* ====== Filter addresses ========================================= */
   addresses = sctp_getpaddrs(fd,0,&addressArray);
   if(addresses > 0) {
      LOG_VERBOSE
      fputs("SCTP association local addresses:\n",stdlog);
      for(i = 0;i < addresses;i++) {
         fprintf(stdlog,"%d/%d: ",i + 1,addresses);
         fputaddress((struct sockaddr*)&addressArray[i],false,stdlog);
         fputs("\n",stdlog);
      }
      LOG_END

      addressesValid = filterAddresses(message->PoolElementPtr,addressArray,addresses);

      LOG_VERBOSE
      fputs("Filtered Registration Request: ",stdlog);
      poolElementPrint(message->PoolElementPtr,stdlog);
      LOG_END

      sctp_freepaddrs(addressArray);
   }
   else {
      /* ?????? */
      addressesValid = true;
   }

   if(addressesValid == false) {
      LOG_ERROR
      fputs("No valid addresses for registration\n",stdlog);
      LOG_END
      message->Error = AEC_AUTHORIZATION_FAILURE;
      return(true);
   }


   /* ====== Check, if pool element already exists ======================= */
   existing = poolNamespaceFindPoolElementByIdentifier(
                  nameServer->Namespace,
                  message->PoolElementPtr->Identifier);
   if(existing == NULL) {
      /* ====== Check, if pool already exists ============================ */
      pool = poolNamespaceFindPool(nameServer->Namespace,message->PoolHandlePtr);
      if(pool != NULL) {
         poolAdded = false;

         /* Try to adapt requested policy to pool's policy. */
         if(poolPolicyAdapt(message->PoolElementPtr->Policy,pool->Policy) == false) {
            LOG_ACTION
            fputs("Pool policy inconsistent -> Registration rejected.\n",stdlog);
            LOG_END
            message->Error = AEC_POLICY_INCONSISTENT;
            pool = NULL;
         }
      }

      /* ====== Pool does not exist -> add it. =========================== */
      else {
         poolAdded = true;

         /* Create pool. */
         pool = poolNew(message->PoolHandlePtr,message->PoolElementPtr->Policy);
         if(pool != NULL) {
            poolNamespaceAddPool(nameServer->Namespace,pool);
         }
         else {
            message->Error = AEC_OUT_OF_MEMORY;
            pool           = NULL;
         }
      }


      /* ====== Add pool element to pool ================================= */
      if(pool != NULL) {
         /* Take pool element from ASAPMessage structure. */
         message->PoolElementPtrAutoDelete = false;

         message->PoolElementPtr->HomeENRPServerID = nameServer->ServerID;
         message->PoolElementPtr->Creator          = (void*)client;
         message->PoolElementPtr->Handle           = getPoolElementHandle();
         poolAddPoolElement(pool,message->PoolElementPtr);
         poolNamespaceAddPoolElement(nameServer->Namespace,message->PoolElementPtr);
         client->PoolElementList = g_list_append(client->PoolElementList,message->PoolElementPtr);

         client->Registrations++;
         if(client->Registrations == 1) {
            LOG_ACTION
            fputs("Starting KeepAlive timers\n",stdlog);
            LOG_END
            timerStart(client->KeepAliveTimer,nameServer->KeepAliveInterval);
            timerStart(client->TimeoutTimer,nameServer->KeepAliveTimeout);
         }


         LOG_ACTION
         fputs("Registration successful!\n",stdlog);
         LOG_END
      }
   }


   /* ====== Pool element already exists -> try re-registration ========== */
   else {
      pool = existing->OwnerPool;

      /* Existing pool element has been created by another client.
         Probably non-unique PE-ID. */
      if(existing->Creator != client) {
         message->Error = AEC_NONUNIQUE_PE_ID;

         LOG_ACTION
         fputs("Non-unique pool element identifier -> Registration rejected.\n",stdlog);
         LOG_END
      }

      /* Re-registration */
      else {
         /* Try to adapt requested policy to pool's policy. */
         if(poolPolicyAdapt(message->PoolElementPtr->Policy,pool->Policy) == false) {
            message->Error = AEC_POLICY_INCONSISTENT;
            LOG_ACTION
            fputs("Pool policy for re-registration inconsistent -> Registration rejected.\n",stdlog);
            LOG_END
         }

         /* Update pool element's settings. */
         else {
            message->Error = poolElementAdapt(existing,message->PoolElementPtr);
            LOG_ACTION
            fputs("Re-registration successful!\n",stdlog);
            nameServerDump(nameServer);
            LOG_END
         }
      }
   }

   return(true);
}


/* ###### Handle deregistration request ################################## */
static bool nameServerDeregistrationRequest(struct NameServer*  nameServer,
                                            struct ENRPClient*  client,
                                            int                 fd,
                                            struct ASAPMessage* message)
{
   struct PoolElement* poolElement;

   LOG_ACTION
   fprintf(stdlog,"Deregistration Request for pool element $%08x\n",message->Identifier);
   LOG_END

   message->Type  = AHT_DEREGISTRATION_RESPONSE;
   message->Error = AEC_OKAY;


   /* ====== Check, if pool element still exists ========================= */
   poolElement = poolNamespaceFindPoolElementByIdentifier(
                    nameServer->Namespace, message->Identifier);
   if(poolElement != NULL) {

      /* ====== Pool element may only be removed by its creator ========== */
      if(poolElement->Creator == client) {
         message->PoolElementPtr = poolElementDuplicate(poolElement);
         nameServerRemovePoolElement(nameServer,client,poolElement);

         client->Registrations--;
         if(client->Registrations == 0) {
            LOG_ACTION
            fputs("Stopping KeepAlive timers\n",stdlog);
            LOG_END
            timerStop(client->KeepAliveTimer);
            timerStop(client->TimeoutTimer);
         }

         LOG_ACTION
         fputs("Deregistration successful!\n",stdlog);
         nameServerDump(nameServer);
         LOG_END
      }

      else {
         message->Error = AEC_AUTHORIZATION_FAILURE;
         LOG_ACTION
         fputs("Deregistration not by creator -> Deregistration rejected!\n",stdlog);
         nameServerDump(nameServer);
         LOG_END
      }
   }

   /* ====== Pool element does not exist -> okay. ======================== */
   else {
      LOG_ACTION
      fputs("Pool element to deregister is not in the namespace!\n",stdlog);
      LOG_END
   }

   return(true);
}


/* ###### Handle name resolution request ################################# */
static bool nameServerNameResolutionRequest(struct NameServer*  nameServer,
                                            struct ENRPClient*  client,
                                            int                 fd,
                                            struct ASAPMessage* message)
{
   struct Pool* pool;

   LOG_ACTION
   fprintf(stdlog,"Name Resolution Request: ");
   poolHandlePrint(message->PoolHandlePtr,stdlog);
   fputs("\n",stdlog);
   LOG_END

   pool = poolNamespaceFindPool(nameServer->Namespace,message->PoolHandlePtr);
   if(pool != NULL) {
      message->Type              = AHT_NAME_RESOLUTION_RESPONSE;
      message->Error             = 0;
      message->PoolPtrAutoDelete = false;
      message->PoolPtr           = pool;

      LOG_ACTION
      fprintf(stdlog,"Result: ");
      poolPrint(message->PoolPtr,stdlog);
      LOG_END
   }
   else {
      message->Type = AHT_NAME_UNKNOWN;
      message->Error = 0;

      LOG_ACTION
      fputs("Pool is not in the namespace!\n",stdlog);
      LOG_END
   }

   return(true);
}


/* ###### Handle endpoint keepalive acknowledgement ###################### */
static void nameServerEndpointKeepAliveAck(struct NameServer*  nameServer,
                                           struct ENRPClient*  client,
                                           int                 fd,
                                           struct ASAPMessage* message)
{
   struct PoolElement* poolElement;

   LOG_VERBOSE
   fprintf(stdlog,"Endpoint KeepAliveAck for $%08x in ",message->Identifier);
   poolHandlePrint(message->PoolHandlePtr,stdlog);
   fputs("\n",stdlog);
   LOG_END

   poolElement = poolNamespaceFindPoolElementByIdentifier(nameServer->Namespace,message->Identifier);
   if(poolElement != NULL) {
      poolElement->TimeStamp            = getMicroTime();
      poolElement->OwnerPool->TimeStamp = poolElement->TimeStamp;

      timerRestart(client->TimeoutTimer,nameServer->KeepAliveTimeout);
   }
   else {
      LOG_VERBOSE
      fputs("Pool element not found!\n",stdlog);
      LOG_END
   }
}


/* ###### Handle server hunt ############################################# */
static bool nameServerHandleServerHunt(struct NameServer*       nameServer,
                                       struct ENRPClient*       client,
                                       int                      fd,
                                       struct ASAPMessage*      message,
                                       struct sockaddr_storage* address,
                                       socklen_t                addressSize)
{
   char str[256];

   LOG_ACTION
   address2string((struct sockaddr*)address,(char*)&str,sizeof(str),true);
   fputs("ServerHunt",stdlog);
   if(addressSize > 0) {
      fprintf(stdlog," from %s (%d)\n",str,addressSize);
   }
   else {
      fputs("\n",stdlog);
   }
   LOG_END

   message->Type = AHT_SERVER_HUNT_RESPONSE;
   return(true);
}


/* ###### Handle events on multicast sockets ############################# */
static void nameServerHandleMulticastEvent(struct NameServer* nameServer,
                                           struct ENRPSource* source,
                                           int                fd)
{
   struct ASAPMessage*     message;
   struct sockaddr_storage senderAddress;
   socklen_t               senderAddressLength;
   bool                    reply;
   size_t                  replyLength;
   ssize_t                 sent;

   LOG_ACTION
   fputs("Entering Multicast Message Handler\n",stdlog);
   LOG_END

   senderAddressLength = sizeof(senderAddress);
   message = asapMessageReceive(fd, 0, 0, 1024, (struct sockaddr*)&senderAddress, &senderAddressLength);
   while(message != NULL) {
      reply = false;
      switch(message->Type) {
         case AHT_SERVER_HUNT:
            reply = nameServerHandleServerHunt(nameServer, NULL, fd, message,
                                               &senderAddress, senderAddressLength);
          break;
         default:
            LOG_WARNING
            fprintf(stdlog,"Message type $%02x via multicast is unsupported!\n",
                    message->Type);
            LOG_END
          break;
      }

      if(reply) {
         replyLength = asapMessage2Packet(message);
         if(replyLength > 0) {
            switch(((struct sockaddr*)&senderAddress)->sa_family) {
               case AF_INET:
                  sent = sendtoplus(nameServer->ReplyIPv4Socket,
                                    message->Buffer, replyLength, 0,
                                    (struct sockaddr*)&senderAddress,
                                    senderAddressLength,
                                    0x00000000, 0, 0, 0, 0);
                break;
#ifdef HAVE_IPV6
               case AF_INET6:
                  sent = sendtoplus(nameServer->ReplyIPv6Socket,
                                    message->Buffer, replyLength, 0,
                                    (struct sockaddr*)&senderAddress,
                                    senderAddressLength,
                                    0x00000000, 0, 0, 0, 0);
                break;
#endif
               default:
                  sent = -1;
                break;
            }
            if(sent != (ssize_t)replyLength) {
               LOG_WARNING
               logerror("Failed sending response to multicast message");
               LOG_END
            }
         }
         else {
            LOG_ERROR
            fputs("Unable to generate reply to multicast message!\n",stdlog);
            LOG_END
         }
      }

      asapMessageDelete(message);
      message = asapMessageReceive(fd, 0, 0, 1024, (struct sockaddr*)&senderAddress, &senderAddressLength);
   }


   LOG_ACTION
   fputs("Leaving Multicast Message Handler\n",stdlog);
   LOG_END
}


/* ###### Handle events on unicast sockets ############################### */
static void nameServerHandleClientEvent(struct NameServer* nameServer,
                                        struct ENRPClient* client,
                                        int                fd)
{
   struct ASAPMessage*     message;
   struct sockaddr_storage senderAddress;
   socklen_t               senderAddressLength;
   bool                    reply;

   LOG_ACTION
   fputs("Entering Unicast Message Handler\n",stdlog);
   LOG_END

   senderAddressLength = sizeof(senderAddress);
   message = asapMessageReceive(fd, 0, 0, 65536, (struct sockaddr*)&senderAddress, &senderAddressLength);
   if(message == NULL) {
      nameServerRemoveClient(nameServer,client);
   }
   else {
      while(message != NULL) {
         reply = false;
         switch(message->Type) {
            case AHT_ENDPOINT_KEEP_ALIVE:
               puts("KeepAlive request received!");
             break;
            case AHT_ENDPOINT_KEEP_ALIVE_ACK:
               nameServerEndpointKeepAliveAck(nameServer,client,fd,message);
             break;
            case AHT_NAME_RESOLUTION:
               reply = nameServerNameResolutionRequest(nameServer,client,fd,message);
             break;
            case AHT_NAME_RESOLUTION_RESPONSE:
             break;
            case AHT_REGISTRATION:
               reply = nameServerRegistrationRequest(nameServer,client,fd,message);
             break;
            case AHT_DEREGISTRATION:
               reply = nameServerDeregistrationRequest(nameServer,client,fd,message);
             break;
            case AHT_SERVER_HUNT:
               reply = nameServerHandleServerHunt(nameServer,client,fd,message,&senderAddress,senderAddressLength);
             break;
            case AHT_SERVER_HUNT_RESPONSE:
               puts("ServerHunt response received!");
             break;
            default:
               LOG_WARNING
               fprintf(stdlog,"Unsupported message type $%02x!\n",
                       message->Type);
               LOG_END
             break;
         }

         if(reply) {
            if(asapMessageSend(fd, 0, message) == false) {
               LOG_ERROR
               logerror("Sending reply failed. Closing connection");
               LOG_END
               nameServerRemoveClient(nameServer,client);
            }
         }

         asapMessageDelete(message);
         message = asapMessageReceive(fd, 0, 0, 65536, (struct sockaddr*)&senderAddress, &senderAddressLength);
      }
   }

   LOG_ACTION
   fputs("Leaving Unicast Message Handler\n",stdlog);
   LOG_END
}


/* ###### Handle events on sockets ####################################### */
static void nameServerSocketHandler(struct Dispatcher* dispatcher,
                                    int                fd,
                                    int                eventMask,
                                    void*              userData)
{
   struct NameServer* nameServer = (struct NameServer*)userData;
   struct ENRPClient* client;
   struct ENRPSource* source;
   int                newfd;

   client = (struct ENRPClient*)g_tree_lookup(nameServer->ClientTree,(gpointer)fd);
   if(client != NULL) {
      nameServerHandleClientEvent(nameServer,client,fd);
   }
   else {
      source = (struct ENRPSource*)g_tree_lookup(nameServer->SourceTree,(gpointer)fd);
      if(source != NULL) {
         if(!source->IsMulticast) {
            newfd = ext_accept(fd,NULL,0);
            if(newfd >= 0) {
               if(setNonBlocking(fd)) {
                  nameServerAddClient(nameServer,newfd);
               }
               else {
                  ext_close(newfd);
               }
            }
         }
         else {
            nameServerHandleMulticastEvent(nameServer,source,fd);
         }
      }
   }
}


/* ###### Send endpoint keepalive messages ############################### */
static void nameServerKeepAliveTimer(struct Dispatcher* dispatcher,
                                     struct Timer*      timer,
                                     void*              userData)
{
   struct ENRPClient*  client = (struct ENRPClient*)userData;
   struct NameServer*  nameServer   = client->Owner;
   GList*              list;
   struct PoolElement* poolElement;
   struct ASAPMessage* message;

   LOG_VERBOSE
   fprintf(stdlog,"KeepAlive-Timer for socket #%d.\n",client->FD);
   LOG_END

   message = asapMessageNew(NULL,1024);
   if(message != NULL) {
      list = g_list_first(client->PoolElementList);
      while(list != NULL) {
         poolElement = (struct PoolElement*)list->data;

         LOG_VERBOSE2
         fprintf(stdlog,"KeepAlive for pool element $%08x in pool ",poolElement->Identifier);
         poolHandlePrint(poolElement->OwnerPool->Handle,stdlog);
         fputs("\n",stdlog);
         LOG_END

         message->PoolElementPtr = poolElement;
         message->PoolHandlePtr  = poolElement->OwnerPool->Handle;
         message->Type           = AHT_ENDPOINT_KEEP_ALIVE;
         if(asapMessageSend(client->FD,0,message) == false) {
            LOG_ERROR
            fputs("Sending KeepAlive failed!\n",stdlog);
            LOG_END
            return;
         }

         asapMessageClearAll(message);

         list = g_list_next(list);
      }
      asapMessageDelete(message);
   }

   timerStart(timer,nameServer->KeepAliveInterval);
}


/* ###### Handle client timeout ########################################## */
static void nameServerTimeoutTimer(struct Dispatcher* dispatcher,
                                   struct Timer*      timer,
                                   void*              userData)
{
   struct ENRPClient* client = (struct ENRPClient*)userData;
   struct NameServer* nameServer   = client->Owner;

   LOG_ACTION
   fprintf(stdlog,"Timeout-Timer for socket #%d -> removing client\n",client->FD);
   LOG_END

   nameServerRemoveClient(nameServer,client);
}




/* ###### Main program ################################################### */
int main(int argc, char** argv)
{
   struct NameServer*      nameServer;
   struct sockaddr_storage localAddress;
   struct sockaddr_storage multicastAddress;
   int                     result;
   struct timeval          timeout;
   fd_set                  readfdset;
   fd_set                  writefdset;
   fd_set                  exceptfdset;
   int                     sockFD;
   int                     i;
   int                     n;
   int                     protocol;
   int                     on;
   int                     defaultProtocol;
   char*                   address;
   unsigned int            sources;


   nameServer = nameServerNew();
   if(nameServer == NULL) {
      puts("ERROR: Out of memory!");
      exit(1);
   }

   if(argc < 2) {
      printf("Usage: %s {[-local=[protocol,]address:port]} {[-mcast=address:port}]} {-logfile=file|-logappend=file|-logquiet} {-loglevel=level} {-logcolor=on|off}\n",argv[0]);
      exit(1);
   }

   puts("The rsplib Name Server - Version 1.0");
   puts("------------------------------------\n");

   sources         = 0;
   defaultProtocol = IPPROTO_TCP;
   for(i = 1;i < argc;i++) {
      if(!(strncasecmp(argv[i],"-local=",7))) {
         address = (char*)&argv[i][7];
         protocol = defaultProtocol;
         if(!(strncasecmp(address,"TCP",3))) {
            protocol = IPPROTO_TCP;
            address  = (char*)&address[4];
         }
         else if(!(strncasecmp(address,"SCTP",4))) {
            protocol = IPPROTO_SCTP;
            address  = (char*)&address[5];
         }
         if(!string2address(address,&localAddress)) {
            printf("ERROR: Bad local address %s! Use format <[protocol/]address:port>.\n",(char*)&argv[i][7]);
            exit(1);
         }
         sockFD = ext_socket(((struct sockaddr*)&localAddress)->sa_family, SOCK_STREAM, protocol);
         if(sockFD >= 0) {
            if(ext_bind(sockFD,(struct sockaddr*)&localAddress,getSocklen((struct sockaddr*)&localAddress)) != 0) {
               perror("ERROR: Unable to bind socket");
               printf("Used local address %s.\n",address);
               exit(1);
            }
            if(!nameServerAddSource(nameServer,sockFD,false)) {
               printf("ERROR: Unable to use local address %s.\n",address);
               exit(1);
            }
            printf("Added local address %s for protocol %s.\n",
                   address,
                   (protocol == IPPROTO_TCP) ? "TCP" : "SCTP");
            sources++;
         }
      }

      else if(!(strncasecmp(argv[i],"-mcast=",7))) {
         address = (char*)&argv[i][7];
         if(!string2address(address,&multicastAddress)) {
            printf("ERROR: Bad local address %s! Use format <[protocol/]address:port>.\n",(char*)&argv[i][7]);
            exit(1);
         }
         memset((char*)&localAddress,0,sizeof(localAddress));
         ((struct sockaddr*)&localAddress)->sa_family = ((struct sockaddr*)&multicastAddress)->sa_family;
         switch(((struct sockaddr*)&multicastAddress)->sa_family) {
            case AF_INET:
               ((struct sockaddr_in*)&localAddress)->sin_port = ((struct sockaddr_in*)&multicastAddress)->sin_port;
             break;
#ifdef HAVE_IPV6
            case AF_INET6:
               ((struct sockaddr_in6*)&localAddress)->sin6_port = ((struct sockaddr_in6*)&multicastAddress)->sin6_port;
             break;
#endif
            default:
               puts("ERROR: Multicast address must be IPv4 or IPv6!");
             break;
         }
         sockFD = ext_socket(((struct sockaddr*)&localAddress)->sa_family, SOCK_DGRAM, IPPROTO_UDP);
         if(sockFD >= 0) {
            on = 1;
            if(ext_setsockopt(sockFD,SOL_SOCKET,SO_REUSEADDR,&on,sizeof(on)) != 0) {
               perror("setsockopt() for SO_REUSEADDR failed");
               exit(1);
            }
            if(ext_bind(sockFD,(struct sockaddr*)&localAddress,getSocklen((struct sockaddr*)&localAddress)) != 0) {
               perror("Unable to bind socket for multicast reception");
               exit(1);
            }
            if(multicastGroupMgt(sockFD,&multicastAddress,NULL,true) == false) {
               printf("ERROR: Unable to join multicast group for address %s!\n",address);
               exit(1);
            }
            if(!nameServerAddSource(nameServer,sockFD,true)) {
               printf("ERROR: Unable to use local address %s.\n",address);
               exit(1);
            }
            printf("Added multicast address %s.\n",address);
            sources++;
         }
      }

      else if(!(strncmp(argv[i],"-log",4))) {
         if(initLogging(argv[i]) == false) {
            exit(1);
         }
      }
   }
   if(sources == 0) {
      puts("ERROR: No sources given. Exiting!");
      exit(1);
   }
   puts("\n");
#ifndef FAST_BREAK
   installBreakDetector();
#endif


   beginLogging();


   while(!breakDetected()) {
      dispatcherGetSelectParameters(nameServer->StateMachine, &n, &readfdset, &writefdset, &exceptfdset, &timeout);
      timeout.tv_sec  = min(0,timeout.tv_sec);
      timeout.tv_usec = min(250000,timeout.tv_usec);
      result = ext_select(n, &readfdset, &writefdset, &exceptfdset, &timeout);
      if(errno == EINTR) {
         break;
      }
      if(result < 0) {
         perror("select() failed");
         break;
      }
      dispatcherHandleSelectResult(nameServer->StateMachine, result, &readfdset, &writefdset, &exceptfdset);
   }

   nameServerDelete(nameServer);
   finishLogging();
   puts("\nTerminated!");
   return(0);
}
