/*
 *  $Id: serverinfolist.c,v 1.11 2002/08/22 09:37:26 dreibh Exp $
 *
 * RSerPool implementation.
 *
 * Realized in co-operation between Siemens AG
 * and University of Essen, Institute of Computer Networking Technology.
 *
 * Acknowledgement
 * This work was partially funded by the Bundesministerium fr Bildung und
 * Forschung (BMBF) of the Federal Republic of Germany (Frderkennzeichen 01AK045).
 * The authors alone are responsible for the contents.
 *
 * This library is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 2.1 of the License, or (at your option) any later version.
 *
 * You should have received a copy of the GNU Lesser General Public
 * License along with this library; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 * There are two mailinglists available at http://www.sctp.de/rserpool.html
 * which should be used for any discussion related to this implementation.
 *
 * Contact: rsplib-discussion@sctp.de
 *          dreibh@exp-math.uni-essen.de
 *
 * Purpose: Name Server Information List
 *
 */


#include "tdtypes.h"
#include "serverinfolist.h"
#include "utilities.h"
#include "netutilities.h"

#include <ext_socket.h>



/* ###### Constructor #################################################### */
struct ServerInfoList* serverInfoListNew()
{
   struct ServerInfoList* serverInfoList = (struct ServerInfoList*)malloc(sizeof(struct ServerInfoList));
   if(serverInfoList != NULL) {
      serverInfoList->UnicastServers       = 0;
      serverInfoList->UnicastServerList    = NULL;
      serverInfoList->MulticastServers     = 0;
      serverInfoList->MulticastServerList  = NULL;
   }
   return(serverInfoList);
}


/* ###### Destructor ##################################################### */
void serverInfoListDelete(struct ServerInfoList* serverInfoList)
{
   if(serverInfoList != NULL) {
      serverInfoListClear(serverInfoList);
      free(serverInfoList);
   }
}


/* ###### Clear ServerInfoList ########################################### */
void serverInfoListClear(struct ServerInfoList* serverInfoList)
{
   struct ServerInfo* serverInfo;
   GList*             list;

   if(serverInfoList != NULL) {
      while(serverInfoList->UnicastServerList != NULL) {
         list       = g_list_first(serverInfoList->UnicastServerList);
         serverInfo = (struct ServerInfo*)list->data;
         serverInfoDelete(serverInfo);
         serverInfoList->UnicastServerList = g_list_remove(serverInfoList->UnicastServerList,serverInfo);
      }
      serverInfoList->UnicastServers = 0;

      while(serverInfoList->MulticastServerList != NULL) {
         list       = g_list_first(serverInfoList->MulticastServerList);
         serverInfo = (struct ServerInfo*)list->data;
         serverInfoDelete(serverInfo);
         serverInfoList->MulticastServerList = g_list_remove(serverInfoList->MulticastServerList,serverInfo);
      }
      serverInfoList->MulticastServers = 0;
   }
}


/* ###### Purge ServerInfoList ########################################### */
void serverInfoListPurge(struct ServerInfoList* serverInfoList, const card64 timeout)
{
   struct ServerInfo* serverInfo;
   GList*             list;
   card64             now;

   if(serverInfoList != NULL) {
      now = getMicroTime();

      list = g_list_first(serverInfoList->UnicastServerList);
      while(list != NULL) {
         serverInfo = (struct ServerInfo*)list->data;
         if(!(serverInfo->Flags & SF_PERMANENT)) {
            if( (serverInfo->Flags & SF_FAILED) ||
                ((timeout != (card64)-1) && (serverInfo->TimeStamp + timeout < now)) ) {
               serverInfoList->UnicastServerList = g_list_remove(serverInfoList->UnicastServerList,serverInfo);
               serverInfoList->UnicastServers--;
               serverInfoDelete(serverInfo);
               list = g_list_first(serverInfoList->UnicastServerList);
               continue;
            }
         }
         list = g_list_next(list);
      }


      list = g_list_first(serverInfoList->MulticastServerList);
      while(list != NULL) {
         serverInfo = (struct ServerInfo*)list->data;
         if(!(serverInfo->Flags & SF_PERMANENT)) {
            if( (serverInfo->Flags & SF_FAILED) ||
                ((timeout != (card64)-1) && (serverInfo->TimeStamp + timeout < now)) ) {
               serverInfoList->MulticastServerList = g_list_remove(serverInfoList->MulticastServerList,serverInfo);
               serverInfoList->MulticastServers--;
               serverInfoDelete(serverInfo);
               list = g_list_first(serverInfoList->MulticastServerList);
               continue;
            }
         }
         list = g_list_next(list);
      }
   }
}


/* ###### Print ServerInfoList ########################################### */
void serverInfoListPrint(struct ServerInfoList* serverInfoList, FILE* fd)
{
   struct ServerInfo* serverInfo;
   GList*             list;

   if(serverInfoList != NULL) {
      fputs("Unicast Servers:\n",fd);
      list = g_list_first(serverInfoList->UnicastServerList);
      while(list != NULL) {
         serverInfo = (struct ServerInfo*)list->data;
         serverInfoPrint(serverInfo,fd);
         list = g_list_next(list);
      }
      fputs("Multicast Servers:\n",fd);
      list = g_list_first(serverInfoList->MulticastServerList);
      while(list != NULL) {
         serverInfo = (struct ServerInfo*)list->data;
         serverInfoPrint(serverInfo,fd);
         list = g_list_next(list);
      }
   }
   else {
      fputs("(null)\n",fd);
   }
}


/* ###### Add ServerInfo to ServerInfoList ############################### */
bool serverInfoListAddServerInfo(struct ServerInfoList* serverInfoList,
                                 struct ServerInfo*     serverInfo)
{
   bool result = false;

   if((serverInfoList) && (serverInfo)) {
      if(serverInfo->Flags & SF_MULTICAST) {
         if(g_list_find_custom(serverInfoList->MulticastServerList,
                               serverInfo,
                               serverInfoCompareFunc) == NULL) {
            serverInfoList->MulticastServerList = g_list_append(serverInfoList->MulticastServerList,serverInfo);
            serverInfoList->MulticastServers++;
            result = true;

            LOG_VERBOSE5
            fputs("Added Multicast ServerInfo already in list: \n",stdlog);
            serverInfoPrint(serverInfo,stdlog);
            LOG_END
         }
         else {
            LOG_VERBOSE5
            fputs("Multicast ServerInfo already in list: \n",stdlog);
            serverInfoPrint(serverInfo,stdlog);
            LOG_END
         }
      }
      else {
         if(g_list_find_custom(serverInfoList->UnicastServerList,
                               serverInfo,
                               serverInfoCompareFunc) == NULL) {
            serverInfoList->UnicastServerList = g_list_append(serverInfoList->UnicastServerList,serverInfo);
            serverInfoList->UnicastServers++;
            result = true;

            LOG_VERBOSE5
            fputs("Added Unicast ServerInfo already to list: \n",stdlog);
            serverInfoPrint(serverInfo,stdlog);
            LOG_END
         }
         else {
            LOG_VERBOSE5
            fputs("Unicast ServerInfo already in list: \n",stdlog);
            serverInfoPrint(serverInfo,stdlog);
            LOG_END
         }
      }
   }

   return(result);
}


/* ###### Remove ServerInfo fron ServerInfoList ########################## */
void serverInfoListRemoveServerInfo(struct ServerInfoList* serverInfoList,
                                    struct ServerInfo*     serverInfo)
{
   if((serverInfoList) && (serverInfo)) {
      if(serverInfo->Flags & SF_MULTICAST) {
         serverInfoList->MulticastServerList = g_list_remove(serverInfoList->MulticastServerList,serverInfo);
         serverInfoList->MulticastServers--;
      }
      else {
         serverInfoList->UnicastServerList = g_list_remove(serverInfoList->UnicastServerList,serverInfo);
         serverInfoList->UnicastServers--;
      }
   }
}


/* ###### Get static servers from configuration file ##################### */
bool serverInfoListGetStaticEntries(struct ServerInfoList* serverInfoList, const char* fileName)
{
   struct sockaddr_storage  address;
   struct ServerInfo*       serverInfo;
   struct TransportAddress* transportAddress;
   FILE*                    fh;
   char                     buffer[512];
   char                     command[64];
   char                     parameter1[256];
   char                     parameter2[256];
   char                     parameter3[256];
   bool                     hasP3;
   size_t                   position;
   size_t                   line;
   int                      protocol = 0;
   unsigned int             flags    = 0;

   fh = fopen(fileName,"r");
   if(fh != NULL) {
      line = 1;
      while(!feof(fh)) {
         if( (fgets((char*)&buffer,sizeof(buffer),fh)) && (strlen(buffer) > 1) ) {
            buffer[strlen(buffer) - 1] = 0x00;
            position = 0;
            LOG_VERBOSE4
            fprintf(stdlog,"Read string <%s>\n",buffer);
            LOG_END
            if(getNextWord(buffer,(char*)&command,sizeof(command),&position)) {
               if(!(strcmp(command,"nameserver"))) {
                  if(getNextWord(buffer,(char*)&parameter1,sizeof(parameter1),&position) &&
                     getNextWord(buffer,(char*)&parameter2,sizeof(parameter2),&position)) {
                     hasP3 = getNextWord(buffer,(char*)&parameter3,sizeof(parameter3),&position);
                     if(hasP3) {
                        LOG_VERBOSE5
                        fprintf(stdlog,"nameserver <%s> <%s> <%s>\n",parameter1,parameter2,parameter3);
                        LOG_END
                     }
                     else {
                        LOG_VERBOSE5
                        fprintf(stdlog,"nameserver <%s> <%s>\n",parameter1,parameter2);
                        LOG_END
                     }
                     if(string2address(parameter1,&address) == false) {
                        LOG_WARNING
                        fprintf(stdlog,"File %s, line %u, column %u: Bad address <%s>!\n",
                               fileName, (unsigned int)line, (unsigned int)position, parameter1);
                        LOG_END
                     }
                     if(!(strcmp(parameter2,"sctp"))) {
                        protocol = IPPROTO_SCTP;
                        flags    = SF_UNKNOWN|SF_STATIC;
                     }
                     else if(!(strcmp(parameter2,"tcp"))) {
                        protocol = IPPROTO_TCP;
                        flags    = SF_UNKNOWN|SF_STATIC;
                     }
                     else if(!(strcmp(parameter2,"multicast"))) {
                        protocol = IPPROTO_UDP;
                        flags    = SF_STATIC|SF_MULTICAST;
                     }
                     else {
                        LOG_WARNING
                        fprintf(stdlog,"File %s, line %u, column %u: Bad protocol <%s>!\n",
                               fileName, (unsigned int)line, (unsigned int)position, parameter2);
                        LOG_END
                     }

                     if(hasP3) {
                        if(!(strcmp(parameter3,"permanent"))) {
                           flags |= SF_PERMANENT;
                        }
                        else {
                           LOG_WARNING
                           fprintf(stdlog,"File %s, line %u, column %u: Bad option <%s>!\n",
                                  fileName, (unsigned int)line, (unsigned int)position, parameter3);
                           LOG_END
                        }
                     }

                     transportAddress = transportAddressNew(protocol,
                                                            getPort((struct sockaddr*)&address),
                                                            (const struct sockaddr_storage*)&address,
                                                            1);
                     if(transportAddress != NULL) {
                        serverInfo = serverInfoNew(0,transportAddress,flags);
                        if(serverInfo != NULL) {
                           serverInfoListAddServerInfo(serverInfoList,serverInfo);
                        }
                        else {
                           transportAddressDelete(transportAddress);
                        }
                     }
                  }
                  else {
                     LOG_WARNING
                     fprintf(stdlog,"File %s, line %u, column %u: Missing parameters!\n",
                            fileName, (unsigned int)line, (unsigned int)position);
                     LOG_END
                  }
               }
               else if(command[0] == '#') {
               }
               else {
                  LOG_WARNING
                  fprintf(stdlog,"File %s, line %u, column %u: Unknown command <%s>!\n",
                          fileName, (unsigned int)line, (unsigned int)position, command);
                  LOG_END
               }
            }
            else {
               LOG_WARNING
               fprintf(stdlog,"File %s, line %u: Line too long!\n",
                       fileName, (unsigned int)line);
               LOG_END
            }
         }

         line++;
      }

      fclose(fh);
   }
   else {
      LOG_ERROR
      fprintf(stdlog,"Unable to read configuration file %s!\n",fileName);
      LOG_END
      return(false);
   }

   LOG_ACTION
   fprintf(stdlog,"ServerInfoList initialized from configuration file %s:\n",fileName);
   serverInfoListPrint(serverInfoList,stdlog);
   LOG_END

   return(true);
}
